/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"
#include "lxml.h"

#include "configuration.h"
#include "identity.h"
#include "mta.h"
#include "processing.h"
#include "folder.h"
#include "mailbox.h"
#include "foldersview.h"

static XePConfiguration *
Configuration_new(void);

static int
Configuration_version_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_identities_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_identity_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_signatures_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_signature_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_defaults_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_ref_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_toplevel_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_cache_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_folders_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_folder_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_rule_refs_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_rule_ref_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_mailboxes_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_mailbox_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_mtas_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_mta_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_server_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_command_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_vfolders_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_vfolder_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_folder_refs_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_folder_ref_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_condition_refs_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_condition_ref_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_processing_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_foldersview_list_cb(cntree *atree, void **data, void **post_data, int op);

static int
Configuration_register_folders(XePConfiguration *configuration);

static int
Configuration_link_folders(XePConfiguration *configuration);

/* Function definitions */

static XePConfiguration *
Configuration_new(void)
{
  XePConfiguration *configuration;

  configuration = x_alloc(1, XePConfiguration);

  if (configuration == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  else
    {
      configuration->identities = a_Identity_list_new();
      configuration->mtas = a_MTA_list_new();
      configuration->vfolders = a_Folder_list_new();
      configuration->processing = a_Processing_new();
      configuration->foldersviewslist = a_Foldersview_list_new(); 
      configuration->folder_refs = a_Folder_ref_hash_new();
      if (!(configuration->identities &&
	    configuration->mtas &&
	    configuration->vfolders &&
	    configuration->folder_refs))
	{
	  a_Configuration_free(configuration);
	  configuration = NULL;
	}
    }

  return configuration;
}

void
a_Configuration_free(XePConfiguration *configuration)
{
  if (!configuration)
    return;

  a_Identity_list_free(configuration->identities);
  a_MTA_list_free(configuration->mtas);
  a_Processing_free();
  a_Folder_list_free(configuration->vfolders);
  a_Folder_ref_hash_free();
  a_Foldersview_list_free(configuration->foldersviewslist);
  x_free(configuration);    
}

int 
Configuration_create(cntree **ntree, XePConfiguration **configuration, int op)
{
  int r = -1;
  cntree *atree =  NULL;
  XePConfiguration *aconfiguration = NULL;

  LXMLelement *element;
  LXMLelement_callbacks element_cb [] = 
    { /* element,          callback,                    data */
      {"version",      Configuration_version_cb,          NULL},
      {"identities",   Configuration_identities_cb,       NULL},
      {"mtas",         Configuration_mtas_cb,             NULL},
      {"vfolders",     Configuration_vfolders_cb,         NULL},
      {"processing",   Configuration_processing_cb,       NULL},
      {"foldersviews", Configuration_foldersview_list_cb, NULL},
      {NULL, NULL, NULL}
    };
  
  if (op == LXML_Read)
    {
      atree = *ntree;
      aconfiguration = Configuration_new();
    }
  else
    {
      atree = lxml_tree_new();
      aconfiguration = *configuration;
    }

  if (atree)
    {
      element = lxml_tree_get_element(atree);
      if (lxml_element_name(element, "configuration", op))
	{
	  A_LOGGING_ERROR("Configuration creation error");
	  r = -1;
	}
      else
	{
	  element_cb[0].data = NULL;
	  element_cb[1].data = (void **)&(aconfiguration->identities);
	  element_cb[2].data = (void **)&(aconfiguration->mtas);
	  element_cb[3].data = (void **)&(aconfiguration->vfolders);
	  element_cb[4].data = (void **)&(aconfiguration->processing);
	  element_cb[5].data = (void **)&(aconfiguration->foldersviewslist);
	  
	  r = lxml_tree_convert(atree, element_cb, op);
	  
	  if (r == 0 && op == LXML_Read)
	    {
	      Configuration_register_folders(aconfiguration);
	      
	      Configuration_link_folders(aconfiguration);
	    }
	}
    }
  
  if (op == LXML_Read)
    {
      *configuration = aconfiguration;
    }
  else
    {
      *ntree = atree;
    }
  return r;
}

XePConfiguration *
a_Configuration_create_from_tree(cntree *ntree)
{
  XePConfiguration *configuration = NULL;
        
  if (Configuration_create(&ntree, &configuration, LXML_Read) != 0)
    {
      A_LOGGING_ERROR("Configuration could not be parsed.");
    }
  
  a_Foldersview_mailboxes_view_update();

  return configuration;
}

cntree *
a_Configuration_create_tree(XePConfiguration *configuration)
{
  int r;
  cntree *ntree = NULL;

  A_LOGGING_TRACE;
  
  r = Configuration_create(&ntree, &configuration, LXML_Write);
  if (r)
    {
      A_LOGGING_ERROR("Configuration could not be created.");
    }

  return ntree;
}


static int
Configuration_version_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r = -1;
  LXMLelement *element;
  LXMLstore store[] =
    {
      {"number", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  char *number = NULL;
  
  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      number = XEP_CONFIGURATION_FILE_FORMAT_VERSION;
    }

  store[0].dest = (void **)&number;
  if (lxml_element_store_attr(element, store, op))
    {
      A_LOGGING_ERROR("<version> attributes error. "
		      "Configuration file ignored");
    }
  else if (op == LXML_Read && 
	   (!number || strcmp(number, XEP_CONFIGURATION_FILE_FORMAT_VERSION)))
    {
      A_LOGGING_ERROR("Configuration file format expected to be of version %s"
		      "\nbut it is of version %s. Configuration ignored.", 
		      XEP_CONFIGURATION_FILE_FORMAT_VERSION,
		      number);
    }
  else
    r = 0;

  x_free(number);

  return r;
}

static int
Configuration_identities_cb(cntree *atree, void **data, void **data_iter, int op)
{
  LXMLelement *element;
  clist *identities = *((clist **) data);
  LXMLelement_callbacks element_cb [] = 
    {
      {"identity", Configuration_identity_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  if (op == LXML_Read && lxml_element_attr_count(element) != 0)
    {
      A_LOGGING_ERROR("<identities> should not have an attribute. "
		      "Identities ignored.");
      return -1;
    }

  element_cb[0].data = (void **)&identities;

  A_LOGGING_INFO("Parsing <identities>...");
  lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_INFO("<identities> parsed");

  return 0;  
}

static int
Configuration_identity_cb(cntree *atree, void **data, void **data_iter, int op)
{
  clist *identities = *((clist **) data);
  clistiter *iter = *((clistiter **) data_iter);
  char *id = NULL;
  char *name = NULL;
  char *email = NULL;
  XePIdentity *identity;
  LXMLelement *element;
  LXMLstore store [] = 
    {
      {"id", LXML_String, NULL, LXML_Unique},
      {"name", LXML_String, NULL, LXML_Unique},
      {"email", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  LXMLelement_callbacks element_cb [] = 
    {
      {"defaults",  Configuration_defaults_cb,  NULL},
      {"signatures", Configuration_signatures_cb, NULL},
      {"mailboxes",   Configuration_mailboxes_cb,   NULL},
      {NULL, NULL, NULL}
    };
  
  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Read && (identity = a_Identity_new()) == NULL)
    return -1;

  if (op == LXML_Write)
    {
      lxml_element_name(element, "identity", op);
      if ((iter == NULL) && (NULL == (iter = clist_begin(identities))))
	return 0;
			     
      identity = (XePIdentity *) iter->data;

      id = (char *) a_Identity_get_id(identity);
      name = (char *) a_Identity_get_name(identity);
      email = (char *) a_Identity_get_email(identity);
      iter = iter->next;
      *data_iter = iter;
    }
  store[0].dest = (void **)&(id);
  store[1].dest = (void **)&(name);
  store[2].dest = (void **)&(email);

  if (lxml_element_store_attr(element, store, op))
    {
      A_LOGGING_ERROR("<identity> attribute error. Identity ignored");
      goto err;
    }
  
  /* A second way to get attributes is to store them in temp variables
   * filled by lxml_element_store_attr and use them afterwards.
   */

  if (op == LXML_Read)
    {
      a_Identity_set_id(identity, id);
      a_Identity_set_name(identity, name);
      a_Identity_set_email(identity, email);
      x_free(id);
      x_free(name);
      x_free(email);
    }

  element_cb[0].data = (void **)&identity;
  element_cb[1].data = (void **)&identity;
  element_cb[2].data = (void **)&identity;

  A_LOGGING_DEBUG("Parsing <identity>...");

  if (lxml_tree_convert(atree, element_cb, op) == 0)
    {
      if (op == LXML_Read)
	clist_append(identities, identity);
    }
  else
    goto err;

  A_LOGGING_DEBUG("<identity> parsed");
  
  return 0;

 err:
  if (op == LXML_Read)
    a_Identity_free(identity);
  return -1;
}

static int
Configuration_defaults_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePIdentity *identity = *((XePIdentity **) data);
  LXMLelement_callbacks element_cb[] =
    {
      {"draft_ref", Configuration_ref_cb, NULL},
      {"queue_ref", Configuration_ref_cb, NULL},
      {"sent_ref",  Configuration_ref_cb, NULL},
      {"mta_ref",   Configuration_ref_cb, NULL},
      {NULL, NULL, NULL}
    };
  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Read && lxml_element_attr_count(element))
    {
      A_LOGGING_ERROR("<defaults> attribute error. Defaults ignored");
      r = -1;
    }
  else
    {
      element_cb[0].data = (void **)&(identity->draft_ref);
      element_cb[1].data = (void **)&(identity->queue_ref);
      element_cb[2].data = (void **)&(identity->sent_ref);
      element_cb[3].data = (void **)&(identity->mta_ref);

      A_LOGGING_DEBUG("Parsing <defaults>...");
      
      r = lxml_tree_convert(atree, element_cb, op);
      
      A_LOGGING_DEBUG("<defaults> parsed");
    }
  return r;
}

static int
Configuration_ref_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  char **ref = (char **) data;
  char *tmp = NULL;
  LXMLstore store[] = 
  {
    {"ref", LXML_String, NULL, LXML_Unique},
    {NULL, 0, NULL, 0}
  };
  A_LOGGING_TRACE;

  if (op == LXML_Write)
    tmp = *ref;

  store[0].dest = (void **)&tmp;

  element = lxml_tree_get_element(atree);

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r == 0 && tmp)
	*ref = strdup(tmp);
      x_free(tmp);
    }

  return r;
}

static int
Configuration_signatures_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePIdentity *identity = *((XePIdentity **) data);
  LXMLelement_callbacks element_cb[] =
    {
      {"signature", Configuration_signature_cb, NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **)&(identity->signatures);

  A_LOGGING_DEBUG("Parsing <signatures>...");
  r = lxml_tree_convert(atree, element_cb, op);
  if (r)
    {
      A_LOGGING_DEBUG("Error parsing <signatures>. Signatures ignored.");
    }
  A_LOGGING_DEBUG("<signatures> parsed");
  return r;
}

static int
Configuration_signature_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  clist *signatures = *(clist **) data;
  clistiter *iter = *(clistiter **) data_iter;
  char *file = NULL;
  LXMLstore store[] = 
  {
    {"file", LXML_String, NULL, LXML_Unique},
    {NULL, 0, NULL, 0}
  };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if (iter == NULL)
	iter = clist_begin(signatures);
      if (iter == NULL)
	return 0;
      file = (char *) iter->data;
      iter = iter->next;
      *data_iter = iter;
    }

  store[0].dest = (void **) &file;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r == 0 && file)
	clist_append(signatures, file);
      else
	x_free(file);
    }

  return r;
}

static int
Configuration_mailboxes_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePIdentity *identity = *((XePIdentity **) data);
  LXMLelement_callbacks element_cb [] = 
    {
      {"mailbox", Configuration_mailbox_cb, NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);
  
  element_cb[0].data = (void **)&(identity->mailboxes);

  A_LOGGING_INFO("Found mailboxes. Parsing...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_INFO("Finished mailboxes parsing.");

  return r;  
}

static int
Configuration_mailbox_cb(cntree *atree, void **data, void **data_iter, int op)
{
  LXMLelement *element;
  clist *mailboxes = *((clist **) data);
  clistiter *iter = *((clistiter **) data_iter);
  XePMailbox *mailbox = NULL;
  char *type = NULL;
  char *id = NULL;
  char *toplevel = NULL;
  int r = 0;
  LXMLstore store[] =
    {
      {"id",   LXML_String, NULL, LXML_Unique},
      {"type", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  LXMLelement_callbacks element_cb [] =
    {
      {"inbox_ref", Configuration_ref_cb,      NULL}, 
      {"draft_ref", Configuration_ref_cb,      NULL},
      {"queue_ref", Configuration_ref_cb,      NULL},
      {"sent_ref",  Configuration_ref_cb,      NULL},
      {"toplevel",  Configuration_toplevel_cb, NULL},
      {"server",    Configuration_server_cb,   NULL},
      {"folders",   Configuration_folders_cb,  NULL},
      {"cache",     Configuration_cache_cb,    NULL},
      {"flags",     Configuration_cache_cb,    NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if (iter == NULL)
	iter = clist_begin(mailboxes);
      if (iter == NULL)
	return 0;
      mailbox = (XePMailbox *) iter->data;
      iter = iter->next;
      *data_iter = iter;
      type = (char *) a_Mailbox_type_as_string(mailbox->type);
      id = mailbox->id;
      toplevel = (char *) a_Folder_get_path(mailbox->folder);
    }

  store[0].dest = (void **)&id;
  store[1].dest = (void **)&type;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r || !type || !id)
	{
	  A_LOGGING_ERROR("<mailbox> attributes error. mailbox ignored");
	}
      else
	{
	  mailbox = a_Mailbox_new(id);
	  a_Mailbox_set_type(mailbox, a_Mailbox_type_from_string(type));
	  if (mailbox->type == XEP_MAILBOX_UNKNOWN)
	    {
	      A_LOGGING_ERROR("<mailbox> unknown type: %s", type);
	      r = -1;
	    }
	}
      x_free(id);
      x_free(type);
    }
  
  if (r == 0)
    {
      element_cb[0].data = (void **)&(mailbox->inbox_ref);
      element_cb[1].data = (void **)&(mailbox->draft_ref);
      element_cb[2].data = (void **)&(mailbox->queue_ref);
      element_cb[3].data = (void **)&(mailbox->sent_ref);
      element_cb[4].data = (void **)&(toplevel);
      element_cb[5].data = (void **)&(mailbox->server);
      element_cb[6].data = (void **)&mailbox;
      element_cb[7].data = (void **)&(mailbox->cache);
      element_cb[8].data = (void **)&(mailbox->flags);
      
      A_LOGGING_INFO("Parsing <mailbox>...");
      r = lxml_tree_convert(atree, element_cb, op);
      A_LOGGING_INFO("<mailbox> parsed");
    }

  if (op == LXML_Read)
    {
      if (toplevel)
	a_Mailbox_set_toplevel(mailbox, toplevel);
      if (r == 0)
	clist_append(mailboxes, mailbox);
      else
	a_Mailbox_free(mailbox);
    }

  return r;
}

static int
Configuration_toplevel_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  char **path = (char **) data;
  char *tmp = NULL;
  LXMLstore store[] = 
  {
    {"path", LXML_String, NULL, LXML_Unique},
    {NULL, 0, NULL, 0}
  };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      tmp = *path;
    }

  store[0].dest = (void **) &tmp;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r == 0 && tmp)
	*path = strdup(tmp);
      x_free(tmp);
    }

  return r;
}

static int
Configuration_cache_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  char **path = (char **) data;
  char *tmp = NULL;
  LXMLstore store[] = 
  {
    {"path", LXML_String, NULL, LXML_Unique},
    {NULL, 0, NULL, 0}
  };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      tmp = *path;
    }

  store[0].dest = (void **) &tmp;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r == 0 && tmp)
	*path = strdup(tmp);
      x_free(tmp);
    }

  return r;
}

static int
Configuration_folders_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePMailbox *mailbox = *((XePMailbox **) data);
  LXMLelement_callbacks element_cb[] =
    {
      {"folder", Configuration_folder_cb, NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;
  
  element = lxml_tree_get_element(atree);
  
  if (op == LXML_Read && lxml_element_attr_count(element) != 0)
    {
      A_LOGGING_ERROR("<folders> should not have attribtes. "
		      "<folders> ignored");
      return -1;
    }

  element_cb[0].data = (void **)&mailbox;

  A_LOGGING_INFO("Parsing <folders>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_INFO("<folders> parsed");

  return r;
}

static int
Configuration_folder_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePMailbox *mailbox = *((XePMailbox **) data);
  clist *folders = mailbox->folders; 
  clistiter *iter = *((clistiter **) data_iter);
  XePFolder *folder = NULL;
  char *id = NULL;
  char *path = NULL;
  LXMLstore store[] = 
    {
      {"id",   LXML_String, NULL,   LXML_Unique}, /* FIXME: id is not needed */
      {"path", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  LXMLelement_callbacks element_cb[] =
    {
      {"rule_refs", Configuration_rule_refs_cb, NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if (iter == NULL && (iter = clist_begin(folders)) == NULL)
	return 0;
      folder = (XePFolder *) iter->data;
      iter = iter->next;
      *data_iter = iter;
      id = (char *) a_Folder_get_id(folder); /* FIXME: id is not needed */
      path = (char *) a_Folder_get_path(folder);
    }

  store[0].dest = (void **) &id;
  store[1].dest = (void **) &path;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("<folder> attributes error. Folder ignored");
	  x_free(id);
	  x_free(path);
	  return -1;
	}
      folder = a_Folder_new(XEP_FOLDER_TYPE_REAL);
      if (folder == NULL)
	return -1;
      a_Folder_set_mailbox(folder, mailbox);
      a_Folder_set_path_p(folder, path);
      a_Folder_update_id(folder);
      if (strcmp(id, a_Folder_get_id(folder)))
	{
	  A_LOGGING_ERROR("inconsistent folder id %s != %s\n", id, 
			  a_Folder_get_id(folder));
	}
      x_free(id);
    }


  element_cb[0].data = (void **)&folder;
 
  A_LOGGING_DEBUG("Parsing <folder>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<folder> parsed");

  if (op == LXML_Read)
    {
      if (r)
	a_Folder_free(folder);
      else
	clist_append(folders, folder);
    }

  return r;
}

static int
Configuration_rule_refs_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePFolder *folder =  *((XePFolder **) data);
  LXMLelement_callbacks element_cb[] =
    {
      {"rule_ref", Configuration_rule_ref_cb, NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **)&(folder->rule_refs);

  A_LOGGING_DEBUG("Parsing <rule_refs>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<rule_refs> parsed");

  return r;
}

static int
Configuration_rule_ref_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  clist *rule_refs = *((clist **) data);
  clistiter *iter = *((clistiter **) data_iter);
  XePProcessingRef *rule_ref;
  char *ref = NULL;
  int enabled = 1;
  LXMLstore store[] =
    {
      {"ref",     LXML_String,  (void **)&ref,     LXML_Unique},
      {"enabled", LXML_Boolean, (void **)&enabled, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if (iter == NULL && (iter = clist_begin(rule_refs)) == NULL)
	return 0;
      rule_ref = (XePProcessingRef *) iter->data;
      iter = iter->next;
      *data_iter = iter;
      ref = rule_ref->ref;
      enabled = rule_ref->enabled;
    }

  store[0].dest = (void **)&ref;
  store[1].dest = (void **)&enabled;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("<rule_ref> attributes error. Rule ref ignored");
	}
      else if (ref)
	{
	  rule_ref = a_Processing_ref_new();
	  if (rule_ref == NULL)
	    {
	      x_free(ref);
	      r = -1;
	    }
	  else
	    {
	      rule_ref->ref = ref;
	      rule_ref->enabled = enabled;
	      clist_append(rule_refs, rule_ref);
	    }
	}
    }

  return r;
}

static int
Configuration_mtas_cb(cntree *atree, void **data, void **data_iter, int op)
{
  LXMLelement *element;
  clist *mtas = *((clist **) data);
  LXMLelement_callbacks element_cb[] =
    {
      {"mta", Configuration_mta_cb, NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Read && lxml_element_attr_count(element) != 0)
    {
      A_LOGGING_ERROR("<mtas> should not have attribtes. <mtas> ignored");
      return -1;
    }

  element_cb[0].data = (void **)&mtas;

  A_LOGGING_INFO("Parsing <mtas>...");
  lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_INFO("<mtas> parsed.");

  return 0;  
}

static int
Configuration_mta_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;

  LXMLelement *element;
  clist *mtas = *((clist **) data);
  clistiter *iter = *((clistiter **) data_iter);
  XePMTA *mta;
  LXMLstore store[] =
    {
      {"id", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  LXMLelement_callbacks element_cb [] = 
    {
      {"server",  Configuration_server_cb,    NULL},
      {"command", Configuration_command_cb, NULL},
      {NULL, NULL, NULL}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if (iter == NULL && (iter = clist_begin(mtas)) == NULL)
	return 0;
      mta = (XePMTA *) iter->data;
      iter = iter->next;
      *data_iter = iter;
    }
  else
    {
      mta = a_MTA_new();
    }
  
  store[0].dest = (void **)&(mta->id);

  if (op == LXML_Read && lxml_element_attr_count(element) != 1)
    {
      A_LOGGING_ERROR("<mta> should  have attributes. MTA ignored");
      a_MTA_free(mta);
      return -1;
    }

  r = lxml_element_store_attr(element, store, op);
  if (op == LXML_Read && r)
    {
      A_LOGGING_ERROR("<mta> error in attributes. MTA ignored");
      a_MTA_free(mta);
      return -1;
    }

  element_cb[0].data = (void **)&(mta->server);
  element_cb[1].data = (void **)&(mta->command);
  if (op == LXML_Write)
    {
      if (mta->type == XEP_MTA_COMMAND)
	element_cb[0].data = NULL;
      else
	element_cb[1].data = NULL;
    }

  A_LOGGING_INFO("Parsing <mta>...");
  lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_INFO("Parsed <mta>");

  if (op == LXML_Read)
    {
      mta->type = (mta->command ? XEP_MTA_COMMAND : 
		   (mta->server->host ? XEP_MTA_SMTP : XEP_MTA_INVALID));
      clist_append(mtas, mta);
    }
  return 0;  
}

static int
Configuration_server_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePServer **server_p = (XePServer **) data;
  LXMLstore store[] = 
    {
      {"host", LXML_String,     NULL, LXML_Unique},
      {"port", LXML_Int,        NULL, LXML_Unique},
      {"login", LXML_String,    NULL, LXML_Unique},
      {"password", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  store[0].dest = (void **)&(*server_p)->host;
  store[1].dest = (void **)&(*server_p)->port;
  store[2].dest = (void **)&(*server_p)->login;
  store[3].dest = (void **)&(*server_p)->password;

  r = lxml_element_store_attr(element, store, op);
  if (r && op == LXML_Read)
    {
      A_LOGGING_ERROR("<server> error in attributes. server data maybe incomplete");
      return -1;
    }

   return r;
}

static int
Configuration_command_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  char **command_p = (char **) data;
  char *tmp = NULL;
  LXMLstore store [] = 
    {
      {"execute", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  A_LOGGING_TRACE;

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      tmp = *command_p;
    }

  store[0].dest = (void **) &tmp;

  r = lxml_element_store_attr(element, store, op);
  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("smtp command error in attributes. SMTP ignored");
	}
      else
	{
	  if (tmp)
	    *command_p = strdup(tmp);
	}
      x_free(tmp);
    }

  return r;
}

static int
Configuration_vfolders_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  clist *vfolders = *((clist **) data);
  LXMLelement_callbacks element_cb [] = 
    { /* element,          callback,                    data */
      {"vfolder",      Configuration_vfolder_cb, NULL},
      {NULL, NULL, NULL}
    };

  element_cb[0].data = (void **)&vfolders;

  A_LOGGING_DEBUG("Parsing <vfolders>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<vfolders> parsed");

  return r;  
}

static int
Configuration_vfolder_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  clist *vfolders = *((clist **) data);
  clistiter *iter = *((clistiter **) data_iter);
  XePFolder *vfolder;
  char *id = NULL;
  LXMLelement *element;
  LXMLstore store[] =
    {
      {"id", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  LXMLelement_callbacks element_cb [] = 
    { /* element,            callback,                     data */
      {"folder_refs",    Configuration_folder_refs_cb, NULL},
      {"condition_refs", Configuration_condition_refs_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if (iter == NULL && (iter = clist_begin(vfolders)) == NULL)
	return 0;
      vfolder = (XePFolder *) iter->data;
      iter = iter->next;
      *data_iter = iter;
      id = (char *) a_Folder_get_path(vfolder);
    }
  else
    {
      vfolder = a_Folder_new(XEP_FOLDER_TYPE_VIRTUAL);
      if (vfolder == NULL)
	return -1;
    }

  store[0].dest = (void **) &id;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("<vfolder> attributes error. VFolder ignored");
	  a_Folder_free(vfolder);
	  return 0;
	}
      a_Folder_set_path(vfolder, id);
      a_Folder_update_id(vfolder);
    }

  element_cb[0].data = (void **)&vfolder;
  element_cb[1].data = (void **)&vfolder;

  A_LOGGING_DEBUG("Parsing <vfolder>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<vfolder> parsed");

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("VFolder '%s' error. VFolder ignored.", id);
	  a_Folder_free(vfolder);
	}
      else
	{
	  clist_append(vfolders, vfolder);
	}
      free(id);
    }
  return r;
}
  
static int
Configuration_folder_refs_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  XePFolder *vfolder = *((XePFolder **) data);
  LXMLelement_callbacks element_cb [] = 
    { /* element         callback,                    data */
      {"folder_ref", Configuration_folder_ref_cb, NULL},
      {NULL, NULL, NULL}
    };

  element_cb[0].data = (void **)&vfolder;

  A_LOGGING_DEBUG("Parsing <folder_refs>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<folder_refs> parsed");

  return r;
}

static int
Configuration_folder_ref_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  XePFolder *vfolder = *((XePFolder **) data);
  clist *folder_refs = vfolder->folder_refs;
  clistiter  *iter = *((clistiter **) data_iter);
  XePFolderRef *folder_ref;

  LXMLelement *element;
  LXMLstore store[] = 
    {
      {"ref",     LXML_String,  NULL, LXML_Unique},
      {"enabled", LXML_Boolean, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  if (op == LXML_Write)
    {
      if (iter == NULL && (iter = clist_begin(folder_refs)) == NULL)
	return 0;
      folder_ref = (XePFolderRef *) iter->data;
      iter = iter->next;
      *data_iter = iter;
    }
  else
    {
      folder_ref = a_Folder_ref_new();
      if (folder_ref == NULL)
	return -1;
    }

  element = lxml_tree_get_element(atree);
  
  store[0].dest = (void **)&(folder_ref->ref);
  store[1].dest = (void **)&(folder_ref->enabled);

  r = lxml_element_store_attr(element, store, op);
  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("VFolder '%s' has incorrect <folder_ref>", 
			  a_Folder_get_id(vfolder));
	  a_Folder_ref_free(folder_ref);
	}
      else
	{
	  clist_append(vfolder->folder_refs, folder_ref);
	}
    }
  
  return r;
}

static int
Configuration_condition_refs_cb(cntree *atree, 
				void **data, 
				void **post_data, int op)
{
  int r;
  XePFolder *vfolder = *((XePFolder **) data);
  LXMLelement_callbacks element_cb [] = 
    { /* element         callback,                    data */
      {"condition_ref", Configuration_condition_ref_cb, NULL},
      {NULL, NULL, NULL}
    };

  element_cb[0].data = (void **)&vfolder;

  A_LOGGING_DEBUG("Parsing <condition_refs>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<condition_refs> parsed");

  return r;
}

static int
Configuration_condition_ref_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  XePFolder *vfolder = *((XePFolder **) data);
  clist *condition_refs = vfolder->condition_refs;
  clistiter *iter = *((clistiter **) data_iter);
  XePProcessingRef *condition_ref;

  LXMLelement *element;
  LXMLstore store[] = 
    {
      {"ref",     LXML_String,  NULL, LXML_Unique},
      {"enabled", LXML_Boolean, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  if (op == LXML_Write)
    {
      if (iter == NULL && (iter = clist_begin(condition_refs)) == NULL)
	return 0;
      condition_ref = (XePProcessingRef *) iter->data;
      iter = iter->next;
      *data_iter = iter;
    }
  else
    {
      condition_ref = a_Processing_ref_new();
      if (condition_ref == NULL)
	return -1;
    }

  element = lxml_tree_get_element(atree);
  
  store[0].dest = (void **)&(condition_ref->ref);
  store[1].dest = (void **)&(condition_ref->enabled);

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("VFolder '%s' has incorrect <condition_ref>", 
			  a_Folder_get_id(vfolder));
	  a_Processing_ref_free(condition_ref);
	}
      else
	{
	  clist_append(vfolder->condition_refs, condition_ref);
	}
    }
  
  return r;
}

static int
Configuration_processing_cb(cntree *atree, void **data, void **data_iter, int op)
{
  return a_Processing_create(atree, (XePProcessing **) data, op);
}

static int
Configuration_foldersview_list_cb(cntree *atree, void **data, void **data_iter, int op)
{
  return a_Foldersview_list_create(atree, (XePFoldersViewsList **) data, 
				   op);
}

static int
Configuration_register_folders(XePConfiguration *configuration)
{
  int r = 0;
  clistiter *id_iter;
  clistiter *mb_iter;
  XePIdentity *identity;
  XePMailbox *mailbox;

  A_LOGGING_TRACE;

  for (id_iter = clist_begin(configuration->identities);
       id_iter && !r;
       id_iter = id_iter->next)
    {
      identity = (XePIdentity *) id_iter->data;
      for (mb_iter = clist_begin(identity->mailboxes);
	   mb_iter && !r;
	   mb_iter = mb_iter->next)
	{
	  mailbox = (XePMailbox *) mb_iter->data;
	  r = a_Folder_register_list( mailbox->folders);
	}
    }
  if (r == 0)  
    r = a_Folder_register_list(configuration->vfolders);

  return r;
}

static int
Configuration_link_folders(XePConfiguration *configuration)
{
  int r;
  A_LOGGING_TRACE;

  r = a_Foldersview_list_link_folders(configuration->foldersviewslist);
  return r;
}

